/*
 *	AMD Elan SC520 watchdog driver 0.01
 *	Copyright (c) Cedric LECOURT <clecourt@iliad.fr>, 2002
 * 
 *	based on
 *
 *	Intel 21285 watchdog driver
 *	Copyright (c) Phil Blundell <pb@nexus.co.uk>, 1998
 *
 * 	and on
 *
 *	SoftDog	0.05:	A Software Watchdog Device
 *	(c) Copyright 1996 Alan Cox <alan@redhat.com>, All Rights Reserved.
 * 	
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 */
 
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/smp_lock.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>

/*
 * Define this to stop the watchdog actually rebooting the machine.
 */
#undef ONLY_TESTING

/*
 * Set default watchdog delay
 */
#if defined(CONFIG_SC520TIMEOUT_1S)
#define TIMER_MARGIN	1		/* (secs) Watchdog default delay will be 1s */
#endif
#if defined(CONFIG_SC520TIMEOUT_2S)
#define TIMER_MARGIN	2		/* (secs) Watchdog default delay will be 2s */
#endif
#if defined(CONFIG_SC520TIMEOUT_4S)
#define TIMER_MARGIN	4		/* (secs) Watchdog default delay will be 4s */
#endif
#if defined(CONFIG_SC520TIMEOUT_8S)
#define TIMER_MARGIN	8		/* (secs) Watchdog default delay will be 8s */
#endif
#if defined(CONFIG_SC520TIMEOUT_16S)
#define TIMER_MARGIN	16		/* (secs) Watchdog default delay will be 16s */
#endif
#if defined(CONFIG_SC520TIMEOUT_32S)
#define TIMER_MARGIN	32		/* (secs) Watchdog default delay will be 32s */
#endif

static int soft_margin = TIMER_MARGIN;	/* in seconds */
static int timer_alive;
static int watchdog_activated;
static unsigned char d0;
static unsigned char d1;
static unsigned char d2;
static unsigned char d3;

static __inline void ElanWriteWord(unsigned short Reg, unsigned short DataWord)
{
	isa_writew (DataWord, 0xDF000 + Reg);
}

static __inline unsigned char ElanReadByte(unsigned short Reg)
{
	return isa_readb (0xDF000 + Reg);
}

static void watchdog_ping(void)
{
	/*
	 *	Write special keyed sequence in the MMCR Watchdog register
	 *	So the watchdog timer will be initialized
	 */


#ifdef ONLY_TESTING
	printk("SC520_WATCHDOG:----PING---\n");
#else 
	if (watchdog_activated==0)
	{
		cli();
		ElanWriteWord(0xCB0,0x3333);	
		ElanWriteWord(0xCB0,0xCCCC);	
		/*
		 *	Program Elan's watchdog : Timer is enabled, Reset is enabled
		 *	+ Set Timeout according to user's choice
		 */
		ElanWriteWord(0xCB0,0xC000|(soft_margin<<2));	
		sti();
		watchdog_activated=1;
	} else {
		cli();
		/*
		 *	Reset Elan's current timeout counter
		 */
		ElanWriteWord(0xCB0,0xAAAA);	
		ElanWriteWord(0xCB0,0x5555);	
		sti();
	}
#endif
}

static int watchdog_stop(void)
{
	if(timer_alive) {
#ifndef ONLY_TESTING
		cli();
		/*
		 *	Reset Elan's timer : The watchdog is no more used
		 */
		ElanWriteWord(0xCB0,0x3333);	
		ElanWriteWord(0xCB0,0xCCCC);	
		ElanWriteWord(0xCB0,0x0000);	
		sti();
		printk("SC520_WATCHDOG: Watchdog unactivated\n");
#endif
	}
	return 0;
}

/*
 *	Allow only one person to hold it open
 */
static int watchdog_open(struct inode *inode, struct file *file)
{

	if(timer_alive)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	timer_alive = 1;
	watchdog_activated=0;
#ifdef CONFIG_SC520_TEXT_DIRECTIVES
	printk("SC520_WATCHDOG: Watchdog opened. Write PING to reset/activate it, SIOP to stop it.\n");
#else
	printk("SC520_WATCHDOG: Watchdog opened. First write will activate it.\n");
#endif
	return 0;
}

static int watchdog_release(struct inode *inode, struct file *file)
{
	if(timer_alive)
	{
		timer_alive = 0;
		watchdog_activated=0;
		MOD_DEC_USE_COUNT;
	}
#ifdef CONFIG_SC520_TEXT_DIRECTIVES
	return 0;
#else
	return watchdog_stop();
#endif
}
static ssize_t watchdog_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	int llen = len;
	int lpos = 0;

	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (len)
	{
#ifdef CONFIG_SC520_TEXT_DIRECTIVES
		while (llen)
		{
			/*  Move old characters one step right */
			d3=d2;
			d2=d1;
			d1=d0;
			/* read one character from user-space */
			if (copy_from_user (&d0, data+lpos, 1))
				return -EFAULT;
			llen--;
			lpos++;
#ifdef ONLY_TESTING
			printk("SC520_WATCHDOG: must read %d more characters, d0=%c d1=%c d2=%c d3=%c\n",llen,d0,d1,d2,d3);
#endif
			/* do we recieve PING or STOP ? */
			if((d0=='G')&&(d1=='N')&&(d2=='I')&&(d3=='P')) {
				/*
				 *	Ping the watchdog
			 	 */
				watchdog_ping();
			} else if((d0=='P')&&(d1=='O')&&(d2=='T')&&(d3=='S')) {
				/*
				 *	Stop the watchdog
		 		 */
				watchdog_stop();
			}
		}
		/* data was emptied */
		return len;
#else
		/* in this mode, every character pings the watchdog */
		watchdog_ping();
		return 1;
#endif
	}
	return 0;
}

static int watchdog_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	int i;
	static struct watchdog_info ident=
	{
		0,
		0,
		"AMD Elan SC520 Watchdog"
	};
	switch(cmd)
	{
		default:
			return -ENOTTY;
		case WDIOC_GETSUPPORT:
			i = verify_area(VERIFY_WRITE, (void*) arg, sizeof(struct watchdog_info));
			if (i)
				return i;
			else
				return copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident));
		case WDIOC_GETSTATUS:
		case WDIOC_GETBOOTSTATUS:
			return put_user(0,(int *)arg);
		case WDIOC_KEEPALIVE:
			return 0;
	}
}

static struct file_operations watchdog_fops=
{
	owner:		THIS_MODULE,
	write:		watchdog_write,
	ioctl:		watchdog_ioctl,
	open:		watchdog_open,
	release:	watchdog_release,
};

static struct miscdevice watchdog_miscdev=
{
	WATCHDOG_MINOR,
	"watchdog",
	&watchdog_fops
};

static int __init sc520_watchdog_init(void)
{
	misc_register(&watchdog_miscdev);
	printk("SC520_WATCHDOG: AMD Elan SC520 Watchdog Timer 0.01\n");
	printk("SC520_WATCHDOG: Written by Cedric LECOURT <clecourt@iliad.fr>\n");
#ifdef CONFIG_SC520_TEXT_DIRECTIVES
	printk("SC520_WATCHDOG: Text commands - Timeout : %d sec\n", soft_margin);
#else 
	printk("SC520_WATCHDOG: Timeout : %d sec\n", soft_margin);
#endif
	return 0;
}

static void __exit sc520_watchdog_exit(void)
{
	misc_deregister(&watchdog_miscdev);
}

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Cedric LECOURT <clecourt@iliad.fr>");
MODULE_DESCRIPTION("AMD Elan SC520 Watchdog driver");
MODULE_LICENSE("GPL");

MODULE_PARM(soft_margin,"i");
MODULE_PARM_DESC(soft_margin,"Watchdog timeout, in sec. Valid values are 1, 2, 4, 8, 16, 32.");

module_init(sc520_watchdog_init);
module_exit(sc520_watchdog_exit);
